/*
   Macro          Life
   Author         Carlo.Hogeveen@xs4all.nl
   Compatibility  TSE Pro v2.80h upwards
   Version        1.2.1   -   27 May 2020

   This macro macro implements Conway's well-known game of Life.

   Installation:
      Put this source in TSE's "mac" directory, and compile it.

   Usage:
      Just run the macro.

   Documentation:
      All built-in: run the macro, then read the status line.

   Tip:
      If you make your font smaller and/or maximize your TSE window before
      running Life, then you get a bigger universe.


   History

      1.0.0    3 December 2006
         Initial release.

      1.1.0    4 December 2006
         Added menu's to save/load universes, and to initialize a universe
         with nothing, with random cells, or with some predefined shapes.

      1.2.0    9 December 2006
         Added the Gosper Glider Gun, which eventually shoots itself.
         Significantly optimized the program and menu structure.

      1.2.1    27 May 2020
         No new features:
         Life was made compatible with TSE Pro v4.41.38 upwards.
*/

#ifdef EDITOR_VERSION
#else
   #define EDITOR_VERSION 2800h
#endif

integer show_delay_duration = 54

integer old_id     = 0
integer new_id     = 0
integer life_cols  = 0
integer life_rows  = 0
integer generation = 0
string  mode  [10] = ""

proc show_help_line(string help_line)
   PutStrAttrXY(1,
                life_rows + 1,
                Format("   " + help_line : -1 * Query(ScreenCols)),
                "",
                Color(black ON white))
end

integer proc count_line_neighbours(integer x, integer y)
   integer result = 0
   GotoLine(y)
   if     x == 1
      result = Val(GetText(life_cols,1))+Val(GetText(x,1))+Val(GetText(x+1,1))
   elseif x == life_cols
      result = Val(GetText(x-1,1))+Val(GetText(x,1))+Val(GetText(1,1))
   else
      result = Val(GetText(x-1,1))+Val(GetText(x,1))+Val(GetText(x+1,1))
   endif
   return(result)
end

integer proc count_neighbours(integer x, integer y)
   integer result = 0
   GotoBufferId(old_id)
   if y == 1
      result = count_line_neighbours(x, life_rows)
             + count_line_neighbours(x, y)
             + count_line_neighbours(x, y + 1)
   elseif y == life_rows
      result = count_line_neighbours(x, y - 1)
             + count_line_neighbours(x, y)
             + count_line_neighbours(x, 1)
   else
      result = count_line_neighbours(x, y - 1)
             + count_line_neighbours(x, y)
             + count_line_neighbours(x, y + 1)
   endif
   GotoBufferId(new_id)
   return(result)
end

proc draw_life(string dead, string alive)
   integer x, y
   GotoBufferId(new_id)
   BufferVideo()
   Set(Attr, Color(bright white ON black))
   ClrScr()
   for y = 1 to life_rows
      GotoLine(y)
      for x = 1 to life_cols
         if Val(GetText(x, 1)) == 0
            PutCharXY((x * 2) - 1, y, dead)
         else
            PutCharXY((x * 2) - 1, y, alive)
         endif
      endfor
   endfor
   UnBufferVideo()
end

proc goto_wikipedia()
   integer org_id = GetBufferId()
   integer tmp_id = 0
   string tmp_dir [255] = GetEnvStr("tmp")
   if tmp_dir == ""
      tmp_dir = GetEnvStr("temp")
   endif
   if tmp_dir == ""
      Warn("Error: No directory environment variable TMP or TEMP defined")
      Warn('Work-around: google or wikipedia "conway life" yourself')
   elseif not EditFile(tmp_dir + "\tmp.html", _DONT_PROMPT_)
      Warn("Error: cannot open a file ", tmp_dir, "\tmp.html")
      Warn('Work-around: google or wikipedia "conway life" yourself')
   else
      tmp_id = GetBufferId()
      EmptyBuffer()
      AddLine("<html>")
      AddLine("   <head>")
      AddLine('      <meta http-equiv="refresh" content="3; url=http://en.wikipedia.org/wiki/Conway%27s_Game_of_Life">')
      AddLine("   </head>")
      AddLine("   <body>")
      AddLine("      <br> <br> <br> <br> <br>")
      AddLine('      <a href="http://en.wikipedia.org/wiki/Conway%27s_Game_of_Life">')
      AddLine("         Conway's Game of Life")
      AddLine("      </a>")
      AddLine("   </body>")
      AddLine("</html>")
      if not SaveAs(CurrFilename(), _OVERWRITE_|_DONT_PROMPT_)
         Warn("Error: cannot save a file ", tmp_dir, "\tmp.html")
         Warn('Work-around: google or wikipedia "conway life" yourself')
      else
         Dos(QuotePath(CurrFilename()), _DONT_PROMPT_|_DONT_CLEAR_)
      endif
      GotoBufferId(org_id)
      AbandonFile(tmp_id)
   endif
end

#if EDITOR_VERSION <= 4400h
   integer seed = 1

   integer proc random()
      integer lo, hi, test
      hi = seed  /  127773
      lo = seed mod 127773
      test = 16807 * lo - (2147483647 mod 16807) * hi
      if test > 0
         seed = test
      else
         seed = test + 2147483647
      endif
      return(seed)
   end
#endif

integer proc random_bit()
   return(random() mod 2)
end

proc pause(integer delay_duration)
   integer i = delay_duration
   while i > 0
   and not KeyPressed()
      if i < 18
         Delay(i)
         i = 0
      else
         Delay(18)
         i = i - 18
      endif
   endwhile
end

proc show_delay(integer delay_duration)
   integer i = show_delay_duration
   integer show_color = Color(black ON white)
   PopWinOpen(Query(ScreenCols) / 2 - 15,
              Query(ScreenRows) / 2 -  2,
              Query(ScreenCols) / 2 + 15,
              Query(ScreenRows) / 2 +  2,
              1,
              "Time delay per generation",
              show_color)
   Set(Attr, show_color)
   ClrScr()
   PutStrAttrXY(5,
                2,
                Format(delay_duration / 18,
                       ".",
                       (delay_duration mod 18) * 1000 / 18 :3:"0",
                       "   seconds"),
                "",
                show_color)
   repeat
      Delay(1)
      if KeyPressed()
         i = 0
      else
         i = i - 1
      endif
   until i == 0
   PopWinClose()
end

proc setup_universe(string name)
   integer row = 0
   integer col = 0
   case name
      when "empty"
         generation = 0
         EmptyBuffer()
         for row = 1 to life_rows
            AddLine(Format("":life_cols:"0"), old_id)
            AddLine(Format("":life_cols:"0"), new_id)
         endfor
      when "random"
         #if EDITOR_VERSION <= 4400h
            seed = (GetClockTicks() mod 99999) + 1
         #endif
         EmptyBuffer()
         for row = 1 to life_rows
            AddLine("")
            BegLine()
            for col = 1 to life_cols
               InsertText(Str(random_bit()))
            endfor
         endfor
      when "glider"
         GotoLine  (life_rows / 2 - 1)
         GotoColumn(life_cols / 2 - 1)
         InsertText("111") Down() Left(3)
         InsertText("100") Down() Left(3)
         InsertText("010")
      when "lwss"
         GotoLine  (life_rows / 2 - 2)
         GotoColumn(life_cols / 2 - 2)
         InsertText("01001") Down() Left(5)
         InsertText("10000") Down() Left(5)
         InsertText("10001") Down() Left(5)
         InsertText("11110")
      when "gosper"
         GotoLine  (1)
         GotoColumn(1)
         InsertText("000000000000000000000000100000000000") Down() Left(36)
         InsertText("000000000000000000000010100000000000") Down() Left(36)
         InsertText("000000000000110000001100000000000011") Down() Left(36)
         InsertText("000000000001000100001100000000000011") Down() Left(36)
         InsertText("110000000010000010001100000000000000") Down() Left(36)
         InsertText("110000000010001011000010100000000000") Down() Left(36)
         InsertText("000000000010000010000000100000000000") Down() Left(36)
         InsertText("000000000001000100000000000000000000") Down() Left(36)
         InsertText("000000000000110000000000000000000000") Down() Left(36)
      when "diehard"
         GotoLine  (life_rows / 2 - 1)
         GotoColumn(life_cols / 2 - 4)
         InsertText("00000010") Down() Left(8)
         InsertText("11000010") Down() Left(8)
         InsertText("01000111")
   endcase
   draw_life(".", "*")
end

proc general_help()
   show_help_line("Close the browser to continue ...")
   goto_wikipedia()
   // draw_life(".", "*")
end

proc save_universe()
   SaveAs()
end

proc load_universe()
   EmptyBuffer()
   BegFile()
   InsertFile()
   UnMarkBlock()
end

proc set_mode(string new_mode)
   mode = new_mode
end

menu file_menu()
   history
   "&Save", save_universe(),,"Save the universe"
   "&Load", load_universe(),,"Load a universe"
end

menu initialize_menu()
   history
   "&Empty"                         , setup_universe("empty")
   "&Random"                        , setup_universe("random")
   "&Glider"                        , setup_universe("glider")
   "&Light Weight Space Ship (LWSS)", setup_universe("lwss")
   "G&osper Glider Gun"             , setup_universe("gosper")
   "&Diehard (130 generations)"     , setup_universe("diehard")
end

menu help_menu()
   history
   "&General help", general_help()
   "&About", Warn("Life v1.1.0 released 4 Dec 2006 by Carlo.Hogeveen@xs4all.nl at http://www.xs4all.nl/~hyphen/tse")
end

menu main_menu()
   history
   "&Run"        , set_mode("run")  ,                , "Start a new Life"
   "&File"       , file_menu()      , _MF_DONT_CLOSE_, "Save or Load a universe"
   "&Initialize" , initialize_menu(), _MF_DONT_CLOSE_, "Big Bang?"
   "&Help"       , help_menu()      , _MF_DONT_CLOSE_, "Get help with your life"
   "&Quit"       , set_mode("quit") ,                , "Quit the program"
end

proc Main()
   integer org_id = GetBufferId()
   integer old_attr = Set(Attr, Color(bright white ON black))
   integer old_cursor = Set(Cursor, OFF)
   integer old_insert = Set(Insert, OFF)
   integer x, y, mouse_x, mouse_y
   integer delay_duration = 18
   integer prev_clockticks = 0
   ClrScr()
   life_cols = Query(ScreenCols) / 2
   life_rows = Query(ScreenRows) - 1
   old_id = CreateTempBuffer()
   new_id = CreateTempBuffer()
   setup_universe("empty")
   repeat
      mode = "edit"
      repeat
         draw_life(".", "*")
         show_help_line(Format(generation,
            "   MouseLeft = (De)Select   MouseRight/F10 = Menu"))
         GetKey()
         case Query(Key)
            when <f10>, <RightBtn>
               GotoBufferId(new_id)
               main_menu()
            when <LeftBtn>
               mouse_x = Query(MouseX)
               mouse_y = Query(MouseY)
               if mouse_x mod 2 <> 0
                  GotoBufferId(new_id)
                  GotoLine(mouse_y)
                  GotoColumn((mouse_x / 2) + 1)
                  if GetText(CurrCol(), 1) == "0"
                     InsertText("1")
                  else
                     InsertText("0")
                  endif
               endif
            when <Escape>
               mode = "quit"
         endcase
      until mode <> "edit"
      while mode == "run"
         if KeyPressed()
            GetKey()
            case Query(Key)
               when <f>, <F>
                  if GetClockTicks() - prev_clockticks < show_delay_duration
                     if delay_duration > 0
                        delay_duration = delay_duration - 1
                     endif
                  endif
                  prev_clockticks = GetClockTicks()
                  show_delay(delay_duration)
               when <s>, <S>
                  if GetClockTicks() - prev_clockticks < show_delay_duration
                     delay_duration = delay_duration + 1
                  endif
                  prev_clockticks = GetClockTicks()
                  show_delay(delay_duration)
               when <Escape>
                  mode = "quit"
               otherwise
                  mode = "edit"
            endcase
         else
            repeat
               draw_life(" ", "*")
               show_help_line(Format(generation,
                  "   F = Faster   S = Slower   OtherKey = EditMode"))
               pause(delay_duration)
               GotoBufferId(new_id)
               MarkLine(1, NumLines())
               Copy()
               GotoBufferId(old_id)
               EmptyBuffer()
               Paste()
               UnMarkBlock()
               GotoBufferId(new_id)
               for y = 1 to life_rows
                  GotoLine(y)
                  for x = 1 to life_cols
                     if Val(GetText(x, 1)) == 1
                        if not ((count_neighbours(x, y) - 1) in 2, 3)
                           GotoColumn(x)
                           InsertText("0")
                        endif
                     else
                        if count_neighbours(x, y) == 3
                           GotoColumn(x)
                           InsertText("1")
                        endif
                     endif
                  endfor
               endfor
               generation = generation + 1
            until KeyPressed()
               or mode <> "run"
         endif
      endwhile
   until mode == "quit"
   Set(Attr  , old_attr)
   Set(Cursor, old_cursor)
   Set(Insert, old_insert)
   GotoBufferId(org_id)
   AbandonFile(old_id)
   AbandonFile(new_id)
   UpdateDisplay(_ALL_WINDOWS_REFRESH_)
   PurgeMacro(SplitPath(CurrMacroFilename(), _NAME_))
end

